package edu.cmu.cs.able.probes.RemosProbe;

import siena.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * ProbeHandle is provided for user application to refer to the probe
 * deployed by the application. After a probe is created, user application
 * can use the probe handle to communicate with Probe through ProbeManagerComm
 * and does not need to user ProbeManagerHandle again.
 */
public class ProbeHandle implements Notifiable {
	/** Refer to the siena bus used by application */
	Siena siena;
	/** This is the reference to the application which holds the handle */
	BandwidthProbeConsumer pc;
	/** The probe manager handle that create this probe handle */
	ProbeManagerHandle pmh;

	/** The probe ID that this handle represents */
        ProbeID prbID;
	/** The type of probe that this handle represents */
        String prbType;

	/** Used to store siena event message */
	NotificationBuffer nb;

	/** Determine whether or not to output debugging message */
	boolean debugf = true;

	/** 
	 * Create a new probe handle using according to siena notification,
	 * which generally is a deploy success message.  
	 * @param pc the application that deploys the probe
	 * @param pmh the probe manager handle associated with the application
	 * @param s siena bus used to transmit event message
	 * @param e siena event message
    	 */
	ProbeHandle(BandwidthProbeConsumer pc, ProbeManagerHandle pmh, 
		Siena s, Notification e ) {
	  debugln("ProbeHander(): start ");

	  siena = s;
	  this.pc = pc;
	  this.pmh = pmh;
	  prbID = new ProbeID(e.getAttribute("ProbeID").stringValue());
	  prbType = e.getAttribute("ProbeType").stringValue();

	  nb = new NotificationBuffer();

          Filter f = new Filter();
          f.addConstraint("ProbeType", "RemosProbe");
          f.addConstraint("ProbeID", prbID.toString());
          f.addConstraint("Action", "Sensed");
          f.addConstraint("Direction", "Reply");

          /* publish interest, waiting for any message from this probe */
          try {
              siena.subscribe(f,this);
		
          } catch (Exception ex) {
              System.err.println("error:"+ex.toString());
          }
	   
	  debugln("ProbeHander(): done. ");
	}

	/**
	 * Activate a probe, make it start monitoring 
	 * @param srcIP IP address of the source node of the network link
	 * @param dstIP IP address of the destination node of the network link
	 * @param frequency monitoring frequency
	 * @return true/false, to indicate whether the operation has succeeded
	 */
	public boolean activate(InetAddress srcIP, InetAddress dstIP,
			double frequency) {
	    boolean result;
	    try {
	    	result = sendTriple(srcIP, dstIP, frequency, "Activate");
	    } catch (ProbeException ex) {
		return false;
	    }
	    return result;
 	}

	/**
	 * Change the monitoring parameters setting for the probe. In current
	 * implementation, it has the same semantics with activate()
	 * @param srcIP IP address of the source node of the network link
	 * @param dstIP IP address of the destination node of the network link
	 * @param frequency monitoring frequency
	 * @return true/false, to indicate whether the operation has succeeded
	 */
	public boolean focus(InetAddress srcIP, InetAddress dstIP,
			     double frequency) {
	    boolean result;
	    try {
	        result = sendTriple(srcIP, dstIP, frequency, "Focus");
	    } catch (ProbeException ex) {
		return false;
	    }
	    return result;
	}

	/**
	 * Not implemented
	 */
	public void query() {
	} 

	/**
	 * Accept the report value from the probe monitoring
	 * @param e the siena event that includes the report value
	 */
	public void onSensed(Notification e) {
	   String str = e.getAttribute("Bandwidth").stringValue();
	   String strID = e.getAttribute("ProbeID").stringValue();
	   debugln("sensed(): by "+strID);
	   Double bw = new Double(str);
	   pc.sensed(bw.doubleValue());
	}

	/**
	 * Siena notification function.
	 * @param e siena event
	 */
	public void notify(Notification e) {
	   if (0 == 
		e.getAttribute("Action").stringValue().compareTo("Sensed"))
		onSensed(e);
	}

	/**
	 * Siena notification function.
	 * @param es siena events
	 */
	public void notify(Notification[] es) { }

	/**
	 * Get a string to indicate the probe ID.
	 * @return return a string which include the probe ID string
	 */
	public String toString() {
	    return new String("this is handle for"+prbID.toString());
	}

	/**
	 * Used by activate() and focus(). Because they have the same semantics
	 * in current implementation, we use this method to do the work for 
	 * them. What it does it to put the siena event message, which includes
	 * the probe monitoring setting information, on the siena bus.
	 * @param ip1 InetAddress of the source node of the network link
	 * @param ip2 InetAddress of the destination node of the network link
	 * @param frequency the monitoring frequency
	 * @param action activate/focus
	 * @return true/false, indicating whether the setting is successful
	 */
	private boolean sendTriple(InetAddress ip1, InetAddress ip2, 
			  double frequency, String action) throws ProbeException {

	    /* Construct the event message */
            Notification e = new Notification();
            e.putAttribute("ProbeType", "RemosProbe");
            e.putAttribute("Action", action);
            e.putAttribute("Direction", "Request");
            e.putAttribute("ProbeID", prbID.toString());
	    e.putAttribute("SourceIP", ip1.getHostAddress());
	    e.putAttribute("DestinationIP", ip2.getHostAddress());
	    e.putAttribute("Frequency", Double.toString(frequency));

            Filter f = new Filter();
            f.addConstraint("ProbeType", "RemosProbe");
	    /* We actually do not use focus here, since focus is the same
	     * with activate in our implementation.
	     * 
	     * TODO:
	     * add the functionality for focus
	     */
            f.addConstraint("Action", "Activate");
            f.addConstraint("Direction", "Reply");

            /* publish request */
            try {
                siena.subscribe(f, nb);
                debugln("subscribe interest");
                siena.publish(e);
                debugln("publish Deploy Request");
            } catch (SienaException ex) {
                System.err.println("error:"+ex.toString());
            }

            /* get reply */
            try {
                Notification reply = nb.getNotification(10000);
                if (reply == null) {
                    debugln("getNotification return null");
                    return false;
                }
		else {
		    String rlt = reply.getAttribute("Result").stringValue();
		    if (rlt.equals("Succeed")) 
			return true;
		    else if (rlt.equals("Fail"))
			return false;
		    else
			throw new ProbeException("Invalide Reply");
		}
	    } catch (Exception ex) {
	    }

	    return false;
	}

	/**
	 * Get the probe ID. Similar with toString(), but toString() has
	 * a prefix string to make it a LITTLE bit understandable.
	 * @return return the probe ID string.
	 */
	public String getID() {
	    return prbID.toString();
	}

	/**
	 * Output debugging message.
	 * @param s the debuggin message
	 */
        private void debugln(String s) {
            System.out.println("RemosProbe.ProbeHandle:"+s);
        }

}
